/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_BATCHNORM_H_
#define AIDGE_CORE_OPERATOR_BATCHNORM_H_

#include <array>
#include <memory>
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {
enum class BatchNormAttr { Epsilon, Momentum };


template <DimIdx_t DIM>
class BatchNorm_Op : public Operator,
                public Registrable<BatchNorm_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const BatchNorm_Op<DIM> &)>,
                public StaticAttributes<BatchNormAttr, float, float> {
public:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, 5> mInputs = {std::make_shared<Tensor>(), std::make_shared<Tensor>(),
                                                      std::make_shared<Tensor>(), std::make_shared<Tensor>(),
                                                      std::make_shared<Tensor>()};
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

   public:
    static constexpr const char *Type = "BatchNorm";

    BatchNorm_Op() = delete;

    using Attributes_ = StaticAttributes<BatchNormAttr, float, float>;
    template <BatchNormAttr e>
    using attr = typename Attributes_::template attr<e>;

    constexpr BatchNorm_Op(float epsilon, float momentum)
        : Operator(Type),
          Attributes_(attr<BatchNormAttr::Epsilon>(epsilon),
                           attr<BatchNormAttr::Momentum>(momentum)),
          mOutput(std::make_shared<Tensor>()) {
        setDatatype(DataType::Float32);
    }

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    BatchNorm_Op(const BatchNorm_Op<DIM>& op)
        : Operator(Type),
          Attributes_(op),
          mOutput(std::make_shared<Tensor>(*op.mOutput))
    {
        // cpy-ctor
        setDatatype(op.mOutput->dataType());
        mImpl = op.mImpl ? Registrar<BatchNorm_Op<DIM>>::create(mOutput->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::BatchNorm_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<BatchNorm_Op<DIM>>(*this);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    constexpr void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 5 && "operators supports only 5 inputs");
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    constexpr void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            for (std::size_t i = nbDataInputs(); i < nbInputs(); ++i) {
                if(mInputs[i]->size() != mInputs[0]->dims()[1]) {
                    mInputs[i]->resize(std::array<DimSize_t, 1>({mInputs[0]->dims()[1]}));
                }
            }
            mOutput->resize(mInputs[0]->dims());
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutput->empty()); }

    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 5 && "operators supports only 5 inputs");
        return *(mInputs[inputIdx].get()); }

    inline Tensor& output(const IOIndex_t /*outputIdx*/) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 5 && "BatchNorm Operators supports only 5 inputs");
        return mInputs[inputIdx];
    }
    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert((outputIdx == 0) && "BatchNorm Operator has only 1 output");
        (void) outputIdx; // avoid unused warning
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 5 && "operators supports only 5 inputs");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }
    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        (void) outputIdx; // avoid unused warning
        return std::static_pointer_cast<Data>(mOutput);
    }


    void setBackend(const std::string &name) {
        mImpl = Registrar<BatchNorm_Op<DIM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[1]->setBackend(name);
        mInputs[2]->setBackend(name);
        mInputs[3]->setBackend(name);
        mInputs[4]->setBackend(name);
    }

    void setDatatype(const DataType &datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[1]->setDatatype(datatype);
        mInputs[2]->setDatatype(datatype);
        mInputs[3]->setDatatype(datatype);
        mInputs[4]->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return 5; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
};

template <DimSize_t DIM>
inline std::shared_ptr<Node> BatchNorm(const float epsilon = 1.0e-5F,
                                       const float momentum = 0.1F,
                                       const std::string& name = "") {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by BatchNorm, not supported");
    auto batchNorm = std::make_shared<Node>(std::make_shared<BatchNorm_Op<static_cast<DimIdx_t>(DIM)>>(epsilon, momentum), name);
    addProducer(batchNorm, 1, std::array<DimSize_t,0>({}), "scale");
    addProducer(batchNorm, 2, std::array<DimSize_t,0>({}), "shift");
    addProducer(batchNorm, 3, std::array<DimSize_t,0>({}), "batch_mean");
    addProducer(batchNorm, 4, std::array<DimSize_t,0>({}), "batch_variance");
    return batchNorm;
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::BatchNormAttr>::data[] = { "Epsilon", "Momentum" };
}

#endif //AIDGE_CORE_OPERATOR_BATCHNORM_H_
